/* **********************************************************************************************************************
   Included Header Files
   **********************************************************************************************************************/

#include "VDEngine.h"



/* **********************************************************************************************************************
   Constructor
   **********************************************************************************************************************/

VDEngine::VDEngine() {
    mapMakerRunning = TRUE;
};



/* **********************************************************************************************************************
   Destructor
   **********************************************************************************************************************/

VDEngine::~VDEngine() {
};



/* **********************************************************************************************************************
   Global Variables
   **********************************************************************************************************************/

// these are used to keep track of whether the windows are open or not
HWND vdEngineWindow = NULL;
HWND mapEditorWindow = NULL;
HWND npcEditorWindow = NULL;



/* **********************************************************************************************************************
   WindowProcedure
   ---------------
   Processes messages for the main window.
   **********************************************************************************************************************/

LRESULT CALLBACK WindowProcedure(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam) {
    switch (Msg) {
        case WM_CLOSE:
            // user clicked "X"
            DestroyWindow(hWnd);
            if (hWnd == mapEditorWindow) {
                mapEditorWindow = NULL;
            } else if (hWnd == npcEditorWindow) {
                npcEditorWindow = NULL;
            };
            break;
        case WM_DESTROY:
            // sent by WM_CLOSE, only exit if it's the main window clicking "X"
            if (hWnd == vdEngineWindow) {
                PostQuitMessage(0);
            } else {
                hWnd = NULL;
            };
            break;
        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                // user selected Exit from menu
                case VD_MENU_FILE_EXIT:
                    PostQuitMessage(0);
                    break;
                case VD_MENU_HELP_ABOUT:
                    (void) MessageBox(NULL, "Varkalas D`Lonovan's 2D Game Engine", "VDEngine", MB_OK);
                    break;
                default:
                    break;
            };
            break;
        default:
            break;
    };
    return DefWindowProc(hWnd, Msg, wParam, lParam);
};



/* **********************************************************************************************************************
   Run
   ---
   Main function of the map maker.
   #param  instance  The instance of the program.
   **********************************************************************************************************************/

void VDEngine::Run(HINSTANCE instance) {
    VDMapEditor mapEditor;
    VDNPCEditor npcEditor;
    VDDirect3D graphics;
    VDInput input;
    VDWorld world;
    VDEntity user;
    VDEntity worldOffset;
    MSG message;
    int width, height;
    int choice;
    POINT mousePosition;
    RECT windowDimensions;
    RECT renderDimensions;
    char vdEngineSpriteFileName[500];

    // initialize variables
    width = 640;
    height = 480;
    user.SetCurrentLocation(0, 0);
    worldOffset.SetCurrentLocation(0, 0);
    choice = 0;
    strcpy_s(vdEngineSpriteFileName, "");

    // register all of the windows
    RegisterWindows(instance);

    // create the application window
    vdEngineWindow = CreateVDEngineWindow(instance, VD_ENGINE_CLASS_NAME, width, height);
    if (vdEngineWindow == NULL) {
        VDError("Failed to create the window.", __FILE__, __LINE__);
    };

    // initialize the graphics API
    graphics.InitializeDirect3D(vdEngineWindow);

    // initialize input
    input.InitializeInput(instance);
    input.InitializeKeyboard(vdEngineWindow);
    input.InitializeMouse(vdEngineWindow);

    // main loop
    while (mapMakerRunning) {
        // dispatch incoming sent messages, check the thread message queue, retrieve message
        while (PeekMessage(&message, NULL, 0, 0, PM_REMOVE)) {
            if (message.message == WM_QUIT) {
                return;
            } else if (message.message == WM_COMMAND) {
                // now we shouldn't be checking this here... but it's the best way to open things from the menu
                switch (LOWORD(message.wParam)) {
                    case VD_MENU_FILE_MAP_EDITOR:
                        // load the Map Editor interface
                        if (mapEditorWindow == NULL) {
                            mapEditorWindow = mapEditor.CreateMapEditor(instance, VD_MAP_EDITOR_CLASS_NAME);
                        };
                        break;
                    case VD_MENU_FILE_NPC_EDITOR:
                        // load the NPC Editor interface
                        if (npcEditorWindow == NULL) {
                            npcEditorWindow = npcEditor.CreateNPCEditor(instance, VD_NPC_EDITOR_CLASS_NAME);
                        };
                        break;
                    default:
                        break;
                };
            };
            TranslateMessage(&message);
            DispatchMessage(&message);
        };

        // 1. check for user input
        // 2. run AI
        // 3. move enemies
        // 4. resolve collisions
        // 5. draw graphics
        // 6. play sounds

        // gather input
        input.RetrieveLostInput();
        input.UpdateInput();
        GetCursorPos(&mousePosition);

        // take a look at the current window location
        (void) GetWindowRect(vdEngineWindow, &windowDimensions);

        // adjust the real mouse position accordingly
        int actualMousePositionX = mousePosition.x - GetSystemMetrics(SM_CXBORDER) - windowDimensions.left;
        int actualMousePositionY = mousePosition.y - (GetSystemMetrics(SM_CYBORDER) + GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU)) - windowDimensions.top;

        // VDEngine interface - Save Map button
        if (SendMessage(mapEditor.GetSaveMapButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
            char text[100];
            (void) GetWindowText(mapEditor.GetMapNameEditBox(), text, sizeof(text));
            if (strcmp(text, "") != 0) {
                (void) strcat_s(text, ".vd");
                world.CreateWorld(text);
            } else {
                MessageBox(vdEngineWindow, "No map name was provided.", "VDEngine", MB_OK);
            };
        };

        // interface - Load Map button
        if (SendMessage(mapEditor.GetLoadMapButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
            char text[500];
            OPENFILENAME openFileBox;
            SecureZeroMemory(&openFileBox, sizeof(OPENFILENAME));

            // load an open file box
                      openFileBox.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_NOCHANGEDIR;
                  openFileBox.hwndOwner = mapEditor.GetWindow();
                  openFileBox.lpstrFile = text;
               openFileBox.lpstrFile[0] = '\0';
             openFileBox.lpstrFileTitle = NULL;
                openFileBox.lpstrFilter = "VDEngine Maps\0*.vd\0All Files\0*.*\0";
            openFileBox.lpstrInitialDir = NULL;
                openFileBox.lStructSize = sizeof(OPENFILENAME);
               openFileBox.nFilterIndex = 1;
                   openFileBox.nMaxFile = sizeof(text);
              openFileBox.nMaxFileTitle = NULL;
            BOOL ret = GetOpenFileName(&openFileBox);
            if (ret != 0) {
                (void) strcpy_s(text, openFileBox.lpstrFile);
            };

            // if the user selected a map file
            if (strcmp(text, "") != 0) {
                // load the world
                // TODO:  Need to load in NPC's separately.  That'll involve changing CreateWorld as well as LoadWorld
                world.LoadWorld(text);

                // add the tile sprites
                for (int x = 0; x < world.GetWorldSizeX(); x++) {
                    for (int y = 0; y < world.GetWorldSizeY(); y++) {
                        graphics.AddSprite(world.GetTileImageFileLocation(x, y),
                                           (float) x * 32.0f,
                                           (float) y * 32.0f,
                                           32,
                                           32,
                                           FALSE,
                                           FALSE);
                    };
                };

                // add the NPC sprites
            };
        };

        // Map Editor interface - Load Sprite button
        if (SendMessage(mapEditor.GetLoadSpriteButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
            OPENFILENAME openFileBox;
            SecureZeroMemory(&openFileBox, sizeof(OPENFILENAME));

            // load an open file box
                      openFileBox.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_NOCHANGEDIR;
                  openFileBox.hwndOwner = mapEditor.GetWindow();
                  openFileBox.lpstrFile = vdEngineSpriteFileName;
               openFileBox.lpstrFile[0] = '\0';
             openFileBox.lpstrFileTitle = NULL;
                openFileBox.lpstrFilter = "Sprites\0*.PNG\0All Files\0*.*\0";
            openFileBox.lpstrInitialDir = NULL;
                openFileBox.lStructSize = sizeof(OPENFILENAME);
               openFileBox.nFilterIndex = 1;
                   openFileBox.nMaxFile = sizeof(vdEngineSpriteFileName);
              openFileBox.nMaxFileTitle = NULL;
            BOOL ret = GetOpenFileName(&openFileBox);
            if (ret != 0) {
                (void) strcpy_s(vdEngineSpriteFileName, openFileBox.lpstrFile);

                // load up the sprite
                RECT drawingArea;
                (void) mapEditor.GetTileSpriteDimensions(&drawingArea);
                mapEditor.SetTileSprite(new VDSprite(graphics.GetDirect3DRenderer(), vdEngineSpriteFileName, 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));
            };
        };

        // NPC Editor interface - Load Sprite button
        if (SendMessage(npcEditor.GetLoadSpriteButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
            char text[500];
            OPENFILENAME openFileBox;
            SecureZeroMemory(&openFileBox, sizeof(OPENFILENAME));

            // load an open file box
                      openFileBox.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_NOCHANGEDIR;
                  openFileBox.hwndOwner = npcEditor.GetWindow();
                  openFileBox.lpstrFile = text;
               openFileBox.lpstrFile[0] = '\0';
             openFileBox.lpstrFileTitle = NULL;
                openFileBox.lpstrFilter = "Sprites\0*.PNG\0All Files\0*.*\0";
            openFileBox.lpstrInitialDir = NULL;
                openFileBox.lStructSize = sizeof(OPENFILENAME);
               openFileBox.nFilterIndex = 1;
                   openFileBox.nMaxFile = sizeof(text);
              openFileBox.nMaxFileTitle = NULL;
            BOOL ret = GetOpenFileName(&openFileBox);
            if (ret != 0) {
                (void) strcpy_s(text, openFileBox.lpstrFile);

                // load up the sprite
                RECT drawingArea;
                (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);
                npcEditor.SetNPCSprite(new VDSprite(graphics.GetDirect3DRenderer(), text, 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));
            };
        };

        // see if any keys were pressed
        if (input.GetKeyboardInput()[DIK_ESCAPE] & 0x80) {
            mapMakerRunning = FALSE;
        } else if (input.GetKeyboardInput()[DIK_LEFT] & 0x80) {
            choice = 1;
        } else if (input.GetKeyboardInput()[DIK_RIGHT] & 0x80) {
            choice = 2;
        } else if (input.GetKeyboardInput()[DIK_UP] & 0x80) {
            choice = 3;
        } else if (input.GetKeyboardInput()[DIK_DOWN] & 0x80) {
            choice = 4;
        };

        // user decided to move
        if (choice >= 1 && choice <= 4) {
            // update the user's location
            user.Move(choice);
            choice = 0;
        };

        // cursor is on the map area
        if (actualMousePositionX >= 0 && actualMousePositionY < width) {
            // user clicks on the map
            if (input.GetMouseInput().rgbButtons[0] & 0x80) {
                // a sprite is loaded
                if (strcmp(vdEngineSpriteFileName, "") != 0) {
                    int worldX = 0;
                    int worldY = 0;

                    // get the current coordinates relative to the user's position (when they move with thearrow keys)
                    float onScreenX = (float) user.GetCurrentLocationX() * 32.0f + actualMousePositionX - (actualMousePositionX % 32);
                    float onScreenY = (float) user.GetCurrentLocationY() * 32.0f + actualMousePositionY - (actualMousePositionY % 32);

                    // set the world coordinates in relation to the screen coordinates
                    if (world.GetWorldSizeX() == 0 && world.GetWorldSizeY() == 0) {
                        // first tile of the map, set the offset of the world coordinates
                        worldOffset.SetCurrentLocation((int) onScreenX / 32, (int) onScreenY / 32);
                    } else {
                        worldX = (int) onScreenX / 32 - worldOffset.GetCurrentLocationX();
                        worldY = (int) onScreenY / 32 - worldOffset.GetCurrentLocationY();
                    };

                    // TODO: REMOVE
                    graphics.RemoveTexts();
                    char testess[100];
                    sprintf_s(testess, "World Size: %d, %d\n", world.GetWorldSizeX(), world.GetWorldSizeY());
                    graphics.AddText(testess, 10, 10, 80, 18, D3DCOLOR_ARGB(255, 255, 255, 255));
                    char worldCoord[100];
                    sprintf_s(worldCoord, "World Position: %d, %d\n", worldX, worldY);
                    graphics.AddText(worldCoord, 10, 20, 80, 38, D3DCOLOR_ARGB(255, 255, 255, 255));

                    // see if the tile is out of the current world bounds
                    if (world.CheckNewLocation(0, worldX, worldY) == NULL) {
                        // if so, reallocate the world
                        int method = world.ReallocateWorld(worldX, worldY);

                        // we have to make sure the world view is offset
                        switch (method) {
                            case 1:
                                worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX() + worldX,
                                                               worldOffset.GetCurrentLocationY());
                                break;
                            case 2:
                                worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX(),
                                                               worldOffset.GetCurrentLocationY() + worldY);
                                break;
                            case 3:
                                worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX() + worldX,
                                                               worldOffset.GetCurrentLocationY() + worldY);
                                break;
                            default:
                                break;
                        };
                    };

                    // the tile was in the negative world space, so it's 0 in the new world space
                    if (worldX < 0) {
                        worldX = 0;
                    };
                    if (worldY < 0) {
                        worldY = 0;
                    };

                    // set the tile status
                    if (SendMessage(mapEditor.GetNullTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 0);
                    } else if (SendMessage(mapEditor.GetOpenTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 1);
                    } else if (SendMessage(mapEditor.GetImpassableTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 2);
                    } else if (SendMessage(mapEditor.GetNewAreaTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 3);
                    } else {
                        world.SetTileStatus(worldX, worldY, 0);
                    };

                    // set the name of the texture in world
                    world.SetTileImageFileLocation(worldX, worldY, vdEngineSpriteFileName);

                    // check if there's a tile already in that spot and remove it if there is
                    graphics.RemoveTileSprite(onScreenX, onScreenY);

                    // add the new sprite
                    graphics.AddSprite(vdEngineSpriteFileName, onScreenX, onScreenY, 32, 32, FALSE, FALSE);
                };
            };
        };

        // draw graphics
        renderDimensions.left = 0;
        renderDimensions.top = 0;
        renderDimensions.right = width - GetSystemMetrics(SM_CXBORDER);
        renderDimensions.bottom = height - GetSystemMetrics(SM_CYBORDER);
        graphics.Render(&renderDimensions, 0, 0, user.GetCurrentLocationX(), user.GetCurrentLocationY());

        // draw Map Editor sprites
        if (mapEditor.GetTileSprite() != NULL && mapEditorWindow != NULL) {
            RECT drawingArea;
            (void) mapEditor.GetTileSpriteDimensions(&drawingArea);
            graphics.Render(&drawingArea, mapEditor.GetWindow(), mapEditor.GetTileSprite());
        };

        // draw NPC Editor sprites
        if (npcEditor.GetNPCSprite() != NULL && npcEditorWindow != NULL) {
            RECT drawingArea;
            (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);
            graphics.Render(&drawingArea, npcEditor.GetWindow(), npcEditor.GetNPCSprite());
        };
    };
};



/* **********************************************************************************************************************
   RegisterWindows
   ---------------
   Registers the window classes that we'll be using in this application.
   #param  instance   Instance of the application.
   **********************************************************************************************************************/

void VDEngine::RegisterWindows(HINSTANCE instance) {
    WNDCLASSEX windowClassStructure;
    ATOM registerClassEx;

    // main window
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_ENGINE_CLASS_NAME;             // class name
    windowClassStructure.lpszMenuName = MAKEINTRESOURCE(VD_MENU);          // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };
    (void) SecureZeroMemory(&windowClassStructure, sizeof(windowClassStructure));

    // Map Editor
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_MAP_EDITOR_CLASS_NAME;         // class name
    windowClassStructure.lpszMenuName = NULL;                              // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };

    // NPC Editor
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_NPC_EDITOR_CLASS_NAME;         // class name
    windowClassStructure.lpszMenuName = NULL;                              // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };
};



/* **********************************************************************************************************************
   CreateVDEngineWindow
   --------------------
   Creates the main window of the engine.
   #param  instance   Instance of the application.
   #param  className  The name of the window class.
   #param  width      Width of the window.
   #param  height     Height of the window.
   #return  A handle to the window.
   **********************************************************************************************************************/

HWND VDEngine::CreateVDEngineWindow(HINSTANCE instance, LPCSTR className, int width, int height) {
    // create the menu resource
    HMENU menu = LoadMenu(instance, MAKEINTRESOURCE(VD_MENU));
    if (menu == NULL) {
        VDError("Failed to create the menu resource.", __FILE__, __LINE__);
    };

    return CreateWindowEx(NULL,                                                   // extended window styles
                          className,                                              // class name
                          "VDEngine",                                             // window name
                          WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_VISIBLE,  // window styles
                          0,                                                      // x coordinate (0 = left)
                          0,                                                      // y coordinate (0 = top)
                          width,                                                  // width
                          height,                                                 // height
                          NULL,                                                   // parent/owner window
                          menu,                                                   // menu
                          instance,                                               // instance
                          NULL);                                                  // used for CREATESTRUCT structure
};



/*  TODO:  I haven't been able to get this to return/set the param to the right fileName!

LPSTR VDEngine::OpenSprite(HWND window, char *fileName) {
    OPENFILENAME openFileBox;

    SecureZeroMemory(&openFileBox, sizeof(OPENFILENAME));

           openFileBox.dwReserved;
                openFileBox.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
              openFileBox.FlagsEx;
            openFileBox.hInstance;
            openFileBox.hwndOwner = window;
            openFileBox.lCustData;
             openFileBox.lpfnHook;
    openFileBox.lpstrCustomFilter;
          openFileBox.lpstrDefExt;
            openFileBox.lpstrFile = fileName;
         openFileBox.lpstrFile[0] = '\0';
       openFileBox.lpstrFileTitle = NULL;
          openFileBox.lpstrFilter = "Sprites\0*.PNG\0All Files\0*.*\0";
      openFileBox.lpstrInitialDir = NULL;
           openFileBox.lpstrTitle;
       openFileBox.lpTemplateName;
          openFileBox.lStructSize = sizeof(OPENFILENAME);
       openFileBox.nFileExtension;
          openFileBox.nFileOffset;
         openFileBox.nFilterIndex = 1;
       openFileBox.nMaxCustFilter;
             openFileBox.nMaxFile = sizeof(fileName);
        openFileBox.nMaxFileTitle = NULL;
           openFileBox.pvReserved;

    GetOpenFileName(&openFileBox);

    //(void) strcpy_s(fileName, sizeof(*fileName), openFileBox.lpstrFile);
    //fileName = openFileBox.lpstrFile;
    return openFileBox.lpstrFile;
};*/